#include <stdlib.h>
#include <stdio.h>
#include <base/system.h>
#include <engine/e_if_other.h>
#include <engine/e_config.h>
#include "es_save.h"


static NETSOCKET sock;

void save_init_connection() {
	NETADDR addr;
	if (net_host_lookup("localhost", &addr, NETTYPE_IPV4) != 0) {
		dbg_msg("save", "cannot find 'localhost'");
		exit(1);
	}
	addr.port = 8400;
	
	sock = net_tcp_create(&addr);
	if (sock == NETSOCKET_INVALID) {
		dbg_msg("save", "cannot create TCP socket");
		exit(1);
	}
	
	if (net_tcp_connect(sock, &addr) != 0) {
		dbg_msg("save", "cannot get connected to scores server");
		exit(1);
	}
	
	dbg_msg("save", "connected to scores server");
}

int save_get_day_score(const char * login, const char * type) {
	char buf[1024];
	
	str_format(buf, sizeof(buf), "get_day_score\n%s\n%s\n", type, login);
	if (net_tcp_send(sock, buf, str_length(buf)) < 0) {
		dbg_msg("save", "cannot send 'get_day_score' message");
		exit(1);
	}
	
	int n = 0;
	if ((n = net_tcp_recv(sock, buf, sizeof(buf) - 1)) < 0) {
		dbg_msg("save", "cannot get 'get_day_score' result");
		exit(1);
	}
	buf[n] = '\0';
	
	return atoi(buf);
}

int save_get_score(const char * login, const char * type) {
	char buf[1024];
	
	str_format(buf, sizeof(buf), "get_score\n%s\n%s\n", type, login);
	if (net_tcp_send(sock, buf, str_length(buf)) < 0) {
		dbg_msg("save", "cannot send 'get_score' message");
		exit(1);
	}
	
	int n = 0;
	if ((n = net_tcp_recv(sock, buf, sizeof(buf) - 1)) < 0) {
		dbg_msg("save", "cannot get 'get_score' result");
		exit(1);
	}
	buf[n] = '\0';
	
	return atoi(buf);
}

void save_incr_score(const char * login, int to_add, const char * type) {
	char buf[1024];
	
	str_format(buf, sizeof(buf), "incr_score\n%s\n%s\n%d\n", type, login, to_add);
	if (net_tcp_send(sock, buf, str_length(buf)) < 0) {
		dbg_msg("save", "cannot send 'incr_score' message");
		exit(1);
	}
	
	dbg_msg("save", "save_incr_score : %s, %d, %s", login, to_add, type);
}

struct save_login_score save_get_max_day_score(const char * type) {
	struct save_login_score best = {"nameless tee", 0};
	char buf[1024];
	
	str_format(buf, sizeof(buf), "get_max_day_score\n%s\n", type);
	if (net_tcp_send(sock, buf, str_length(buf)) < 0) {
		dbg_msg("save", "cannot send 'get_max_day_score' message");
		exit(1);
	}
	
	int n = 0;
	if ((n = net_tcp_recv(sock, buf, sizeof(buf) - 1)) < 0) {
		dbg_msg("save", "cannot get 'get_max_day_score' result");
		exit(1);
	}
	buf[n] = '\0';
	
	int i;
	for (i = 0; i < sizeof(buf); i++) {
		if (buf[i] == '\n' || i == sizeof(best.login))
			break;
		best.login[i] = buf[i];
	}
	if (i == sizeof(best.login)) {
		dbg_msg("save", "login result too long");
		exit(1);
	}
	best.login[i] = '\0';
	sscanf(buf + i, "%d", &best.score);
	
	return best;
}

struct save_login_score save_get_max_score(const char * type) {
	struct save_login_score best = {"nameless tee", 0};
	char buf[1024];
	
	str_format(buf, sizeof(buf), "get_max_score\n%s\n", type);
	if (net_tcp_send(sock, buf, str_length(buf)) < 0) {
		dbg_msg("save", "cannot send 'get_max_score' message");
		exit(1);
	}
	
	int n = 0;
	if ((n = net_tcp_recv(sock, buf, sizeof(buf) - 1)) < 0) {
		dbg_msg("save", "cannot get 'get_max_score' result");
		exit(1);
	}
	buf[n] = '\0';
	
	int i;
	for (i = 0; i < sizeof(buf); i++) {
		if (buf[i] == '\n' || i == sizeof(best.login))
			break;
		best.login[i] = buf[i];
	}
	if (i == sizeof(best.login)) {
		dbg_msg("save", "login result too long");
		exit(1);
	}
	best.login[i] = '\0';
	sscanf(buf + i, "%d", &best.score);
	
	return best;
}

