#include <stdio.h>
#include <string.h>
#include <new>
#include <engine/e_config.h>
#include <engine/e_server_interface.h>
#include "gamecontext.hpp"

GAMECONTEXT game;

GAMECONTEXT::GAMECONTEXT()
{
	for(int i = 0; i < MAX_CLIENTS; i++)
		players[i] = 0;
	
	controller = 0;
	vote_closetime = 0;
}

GAMECONTEXT::~GAMECONTEXT()
{
	for(int i = 0; i < MAX_CLIENTS; i++)
		delete players[i];
}

void GAMECONTEXT::clear()
{
	this->~GAMECONTEXT();
	mem_zero(this, sizeof(*this));
	new (this) GAMECONTEXT();
}


class CHARACTER *GAMECONTEXT::get_player_char(int client_id)
{
	if(client_id < 0 || client_id >= MAX_CLIENTS || !players[client_id])
		return 0;
	return players[client_id]->get_character();
}

void GAMECONTEXT::create_damageind(vec2 p, float angle, int amount)
{
	float a = 3 * 3.14159f / 2 + angle;
	//float a = get_angle(dir);
	float s = a-pi/3;
	float e = a+pi/3;
	for(int i = 0; i < amount; i++)
	{
		float f = mix(s, e, float(i+1)/float(amount+2));
		NETEVENT_DAMAGEIND *ev = (NETEVENT_DAMAGEIND *)events.create(NETEVENTTYPE_DAMAGEIND, sizeof(NETEVENT_DAMAGEIND));
		if(ev)
		{
			ev->x = (int)p.x;
			ev->y = (int)p.y;
			ev->angle = (int)(f*256.0f);
		}
	}
}

void GAMECONTEXT::create_hammerhit(vec2 p)
{
	// create the event
	NETEVENT_HAMMERHIT *ev = (NETEVENT_HAMMERHIT *)events.create(NETEVENTTYPE_HAMMERHIT, sizeof(NETEVENT_HAMMERHIT));
	if(ev)
	{
		ev->x = (int)p.x;
		ev->y = (int)p.y;
	}
}


void GAMECONTEXT::create_explosion(vec2 p, int owner, int weapon, bool bnodamage, int damage)
{
	// create the event
	NETEVENT_EXPLOSION *ev = (NETEVENT_EXPLOSION *)events.create(NETEVENTTYPE_EXPLOSION, sizeof(NETEVENT_EXPLOSION));
	if(ev)
	{
		ev->x = (int)p.x;
		ev->y = (int)p.y;
	}

	if (!bnodamage)
	{
		// deal damage
		CHARACTER *ents[64];
		float radius = 135.0f;
		float innerradius = 48.0f;
		int num = game.world.find_entities(p, radius, (ENTITY**)ents, 64, NETOBJTYPE_CHARACTER);
		for(int i = 0; i < num; i++)
		{
			vec2 diff = ents[i]->pos - p;
			vec2 forcedir(0,1);
			float l = length(diff);
			if(l)
				forcedir = normalize(diff);
			l = 1-clamp((l-innerradius)/(radius-innerradius), 0.0f, 1.0f);
			float dmg = damage * l;
			if(l > 0)
				ents[i]->take_damage(forcedir*l*12, (int)dmg, owner, weapon);
		}
	}
}

/*
void create_smoke(vec2 p)
{
	// create the event
	EV_EXPLOSION *ev = (EV_EXPLOSION *)events.create(EVENT_SMOKE, sizeof(EV_EXPLOSION));
	if(ev)
	{
		ev->x = (int)p.x;
		ev->y = (int)p.y;
	}
}*/

void GAMECONTEXT::create_playerspawn(vec2 p)
{
	// create the event
	NETEVENT_SPAWN *ev = (NETEVENT_SPAWN *)events.create(NETEVENTTYPE_SPAWN, sizeof(NETEVENT_SPAWN));
	if(ev)
	{
		ev->x = (int)p.x;
		ev->y = (int)p.y;
	}
}

void GAMECONTEXT::create_death(vec2 p, int cid)
{
	// create the event
	NETEVENT_DEATH *ev = (NETEVENT_DEATH *)events.create(NETEVENTTYPE_DEATH, sizeof(NETEVENT_DEATH));
	if(ev)
	{
		ev->x = (int)p.x;
		ev->y = (int)p.y;
		ev->cid = cid;
	}
}

void GAMECONTEXT::create_sound(vec2 pos, int sound, int mask)
{
	if (sound < 0)
		return;

	// create a sound
	NETEVENT_SOUNDWORLD *ev = (NETEVENT_SOUNDWORLD *)events.create(NETEVENTTYPE_SOUNDWORLD, sizeof(NETEVENT_SOUNDWORLD), mask);
	if(ev)
	{
		ev->x = (int)pos.x;
		ev->y = (int)pos.y;
		ev->soundid = sound;
	}
}

void GAMECONTEXT::create_sound_global(int sound, int target)
{
	if (sound < 0)
		return;

	NETMSG_SV_SOUNDGLOBAL msg;
	msg.soundid = sound;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(target);
}


void GAMECONTEXT::send_chat_target(int to, const char *text)
{
	NETMSG_SV_CHAT msg;
	msg.team = 0;
	msg.cid = -1;
	msg.message = text;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(to);
}


void GAMECONTEXT::send_chat(int chatter_cid, int team, const char *text)
{
	// manage chat commands
	// /info
	if(!str_comp_nocase(text, "/info"))
	{
		if(config.sv_saved_scores)
		{
			send_chat_target(chatter_cid, "--- MODsave ---");
			send_chat_target(chatter_cid, "- All your kills are saved for your login.");
			send_chat_target(chatter_cid, "- There are standard mod and instagib mod scores databases.");
			send_chat_target(chatter_cid, "- You can try to be the Day Tee (best Tee of the day) or even the Tee God (best Tee) !");
			send_chat_target(chatter_cid, "- Commands : /info, /cmdlist, /me, /stat.");
		}
		else
		{
			char buf[1024];
			if(config.sv_seen_gametype[0])
				str_format(buf, sizeof(buf), "--- %s ---", config.sv_seen_gametype);
			else
				str_format(buf, sizeof(buf), "--- %s ---", config.sv_gametype);
			send_chat_target(chatter_cid, buf);
			send_chat_target(chatter_cid, "By Clarus.");
		}
		return;
	}
	
	// /cmdlist
	if(config.sv_saved_scores && !str_comp_nocase(text, "/cmdlist"))
	{
		send_chat_target(chatter_cid, "--- Command list ---");
		send_chat_target(chatter_cid, "/info : get mod description");
		send_chat_target(chatter_cid, "/me : show your scores to other players");
		send_chat_target(chatter_cid, "/stat : see your stats");
		return;
	}
	
	// /stat
	if(config.sv_saved_scores && !str_comp_nocase(text, "/stat"))
	{
		send_chat_precise_stats(chatter_cid);
		return;
	}
	
	// /me
	if(config.sv_saved_scores && !str_comp_nocase(text, "/me"))
	{
		send_broadcast_stats(chatter_cid);
		return;
	}

	NETMSG_SV_CHAT msg;
	if (team == CHAT_ALL)
		msg.team = 0;
	else
		msg.team = 1;
	
	msg.cid = chatter_cid;
	
	// protection contre le passage de numéros ICQ, discussion sur les bots, etc ...
	bool filtered = false;
	
	for(int i = 0; i < MAX_CLIENTS; i++)
		if (game.players[i] && (team == CHAT_ALL || game.players[i]->team == team)) {
			if (config.sv_censorship && i != chatter_cid && chatter_cid != -1) {
				if (str_find_nocase(text, ".de") || str_find_nocase(text, "www") ||
					str_find_nocase(text, "@") || str_find_nocase(text, "icq") ||
					str_find_nocase(text, "tee-bot") || str_find_nocase(text, "home") ||					
					str_find_nocase(text, "msn") || str_find_nocase(text, "mail") ||
					str_find_nocase(text, "skype") || str_find_nocase(text, "dbg") ||
					str_length(text) == 1) {
					filtered = true;
					continue;
				}
				
				int l = str_length(text);
				char text_copy[1000];
				str_copy(text_copy, text, sizeof(text_copy));
				msg.message = text_copy;
				
				int n_numbers = 0;
				for (int i = 0; i < l; i++)
					if ('0' <= text[i] && text[i] <= '9')
						n_numbers++;
				
				if (n_numbers >= 7) {
					filtered = true;
					for (int i = 0; i < l; i++)
						if ('0' <= text[i] && text[i] <= '9')
							text_copy[i] = '0' + (rand() % 10);
				}
			} else
				msg.message = text;
			
			msg.pack(MSGFLAG_VITAL);
			server_send_msg(i);
		}
	
	if (chatter_cid >= 0 && chatter_cid < MAX_CLIENTS) {
		if (filtered)
			dbg_msg("chat", "%d:%d:%s: %s (filtered)", chatter_cid, team, server_clientname(chatter_cid), text);
		else
			dbg_msg("chat", "%d:%d:%s: %s", chatter_cid, team, server_clientname(chatter_cid), text);
	} else
		dbg_msg("chat", "*** %s", text);
}

void GAMECONTEXT::send_emoticon(int cid, int emoticon)
{
	NETMSG_SV_EMOTICON msg;
	msg.cid = cid;
	msg.emoticon = emoticon;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(-1);
}

void GAMECONTEXT::send_weapon_pickup(int cid, int weapon)
{
	NETMSG_SV_WEAPONPICKUP msg;
	msg.weapon = weapon;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(cid);
}


void GAMECONTEXT::send_broadcast(const char *text, int cid)
{
	NETMSG_SV_BROADCAST msg;
	msg.message = text;
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(cid);
}

// send a message with player's stats
void GAMECONTEXT::send_chat_stats(int cid) {
	int score = save_get_score(server_clientname(cid), "MOD");
	int day_score = save_get_day_score(server_clientname(cid), "MOD");
	
	int iscore = save_get_score(server_clientname(cid), "iMOD");
	int day_iscore = save_get_day_score(server_clientname(cid), "iMOD");
	
	char buffer[1000];
	send_chat_target(cid, "Your scores :");
	str_format(buffer, sizeof(buffer),
		"Standard mod : %d kills (%d today)",
		score, day_score);
	send_chat_target(cid, buffer);
	str_format(buffer, sizeof(buffer),
		"Instagib mod : %d kills (%d today)",
		iscore, day_iscore);
	send_chat_target(cid, buffer);
}

// send a message with player's stats and kills to do to be Day Tee or Tee God
void GAMECONTEXT::send_chat_precise_stats(int cid) {
	int score = save_get_score(server_clientname(cid), "MOD");
	int day_score = save_get_day_score(server_clientname(cid), "MOD");
	int best_score = save_get_max_score("MOD").score;
	int best_day_score = save_get_max_day_score("MOD").score;
	
	int iscore = save_get_score(server_clientname(cid), "iMOD");
	int day_iscore = save_get_day_score(server_clientname(cid), "iMOD");
	int best_iscore = save_get_max_score("iMOD").score;
	int best_day_iscore = save_get_max_day_score("iMOD").score;
	
	char buffer[1024];
	send_chat_target(cid, "Your scores :");
	str_format(buffer, sizeof(buffer),
		"Standard mod : %d kills (%d today) (%d kills to be Day Tee, %d kills for Tee God)",
		score, day_score,
		best_day_score - day_score, best_score - score);
	send_chat_target(cid, buffer);
	str_format(buffer, sizeof(buffer),
		"Instagib mod : %d kills (%d today) (%d kills to be Day Tee, %d kills for Tee God)",
		iscore, day_iscore,
		best_day_iscore - day_iscore, best_iscore - iscore);
	send_chat_target(cid, buffer);
}

// brodcast player's stats
void GAMECONTEXT::send_broadcast_stats(int cid) {
	char buffer[1024];
	PLAYER *p = game.players[cid];
	
	if (p) {
		int64 now = time_get();
		int64 time_left = p->last_stats_broadcast + time_freq()*60 - now;
		if (time_left > 0) {
			str_format(buffer, sizeof(buffer), "You must wait %d seconds before making another '/me'.",
				(time_left/time_freq())+1);
			game.send_chat_target(cid, buffer);
			return;
		}
		p->last_stats_broadcast = now;
		
		int score = save_get_score(server_clientname(cid), "MOD");
		int iscore = save_get_score(server_clientname(cid), "iMOD");
		
		str_format(buffer, sizeof(buffer), "'%s' scores :\nStandard mod : %d kills\nInstagib mod : %d kills",
			server_clientname(cid), score, iscore);
		send_broadcast(buffer, -1);
	}
}

void GAMECONTEXT::start_vote(const char *desc, const char *command)
{
	// check if a vote is already running
	if(vote_closetime)
		return;

	// reset votes
	vote_enforce = VOTE_ENFORCE_UNKNOWN;
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(players[i])
			players[i]->vote = 0;
	}
	
	// start vote
	vote_closetime = time_get() + time_freq()*25;
	str_copy(vote_description, desc, sizeof(vote_description));
	str_copy(vote_command, command, sizeof(vote_description));
	send_vote_set(-1);
	send_vote_status(-1);
}


void GAMECONTEXT::end_vote()
{
	vote_closetime = 0;
	send_vote_set(-1);
}

void GAMECONTEXT::send_vote_set(int cid)
{
	NETMSG_SV_VOTE_SET msg;
	if(vote_closetime)
	{
		msg.timeout = (vote_closetime-time_get())/time_freq();
		msg.description = vote_description;
		msg.command = vote_command;
	}
	else
	{
		msg.timeout = 0;
		msg.description = "";
		msg.command = "";
	}
	msg.pack(MSGFLAG_VITAL);
	server_send_msg(cid);
}

void GAMECONTEXT::send_vote_status(int cid)
{
	NETMSG_SV_VOTE_STATUS msg = {0};
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(players[i])
		{
			msg.total++;
			if(players[i]->vote > 0)
				msg.yes++;
			else if(players[i]->vote < 0)
				msg.no++;
			else
				msg.pass++;
		}
	}	

	msg.pack(MSGFLAG_VITAL);
	server_send_msg(cid);
	
}

void GAMECONTEXT::abort_vote_kick_on_disconnect(int client_id)
{
	if(vote_closetime && !strncmp(vote_command, "kick ", 5) && atoi(&vote_command[5]) == client_id)
		vote_closetime = -1;
}

void GAMECONTEXT::tick()
{
	world.core.tuning = tuning;
	world.tick();

	//if(world.paused) // make sure that the game object always updates
	controller->tick();
		
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(players[i])
			players[i]->tick();
	}
	
	// update voting
	if(vote_closetime)
	{
		// abort the kick-vote on player-leave
		if(vote_closetime == -1)
		{
			send_chat(-1, GAMECONTEXT::CHAT_ALL, "Vote aborted");
			end_vote();
		}
		else
		{
			// count votes
			int total = 0, yes = 0, no = 0;
			for(int i = 0; i < MAX_CLIENTS; i++)
			{
				if(players[i])
				{
					total++;
					if(players[i]->vote > 0)
						yes++;
					else if(players[i]->vote < 0)
						no++;
				}
			}
		
			if(vote_enforce == VOTE_ENFORCE_YES || yes >= total/2+1)
			{
				console_execute_line(vote_command);
				end_vote();
				send_chat(-1, GAMECONTEXT::CHAT_ALL, "Vote passed");
			
				if(players[vote_creator])
					players[vote_creator]->last_votecall = 0;
			}
			else if(vote_enforce == VOTE_ENFORCE_NO || time_get() > vote_closetime || no >= total/2+1 || yes+no == total)
			{
				end_vote();
				send_chat(-1, GAMECONTEXT::CHAT_ALL, "Vote failed");
			}
		}
	}
}

void GAMECONTEXT::snap(int client_id)
{
	world.snap(client_id);
	controller->snap(client_id);
	events.snap(client_id);
	
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(players[i])
			players[i]->snap(client_id);
	}
	
	// anti-cheat
	if (checked_player >= 0 && !players[checked_player])
		checked_player = -1;
	if (0 && checked_player >= 0)
	{
		NETOBJ_CLIENT_INFO *client_info = (NETOBJ_CLIENT_INFO *)snap_new_item(NETOBJTYPE_CLIENT_INFO, 15, sizeof(NETOBJ_CLIENT_INFO));
		str_to_ints(&client_info->name0, 6, "");
		str_to_ints(&client_info->skin0, 6, checked_skin);
		client_info->use_custom_color = 0;
		client_info->color_body = 0;
		client_info->color_feet = 0;

		NETOBJ_PLAYER_INFO *info = (NETOBJ_PLAYER_INFO *)snap_new_item(NETOBJTYPE_PLAYER_INFO, 15, sizeof(NETOBJ_PLAYER_INFO));
		info->latency = 0;
		info->latency_flux = 0;
		info->local = 0;
		info->cid = 15;
		info->score = 0;
		info->team = checked_team;
		
		NETOBJ_CHARACTER *character = (NETOBJ_CHARACTER *)snap_new_item(NETOBJTYPE_CHARACTER, 15, sizeof(NETOBJ_CHARACTER));
		character->tick = server_tick();
		character->x = checked_pos.x;
		character->y = checked_pos.y;
		character->vx = 0;
		character->vy = 0;
		character->hook_state = HOOK_IDLE;
		character->hook_tick = 0;
		character->hook_x = 0;
		character->hook_y = 0;
		character->hook_dx = 0;
		character->hook_dy = 0;
		character->hooked_player = 0;
		character->jumped = 0;
		character->direction = 0;
		character->angle = 0;
		character->emote = EMOTE_NORMAL;
		character->ammocount = 0;
		character->health = 0;
		character->armor = 0;
		character->weapon = WEAPON_RIFLE;
		character->attacktick = 0;
		character->direction = 0;
		character->player_state = PLAYERSTATE_PLAYING;
	}
}
