require "rubygems"
require "pg"
require "socket"


class Scores
  attr_accessor :conn, :scores, :max_day_scores, :max_scores
  
  def initialize(base_name)
    @conn = PGconn.open(:dbname => base_name)
    @conn.set_client_encoding("LATIN9")

    @scores = {}
    @max_day_scores = {}
    @max_scores = {}
    res = @conn.exec("SELECT * FROM scores")
    res.each do |line|
      login, score, type = line["login"], line["score"].to_i, line["type"]
      
      @scores[type] = {}  if @scores[type].nil?
      if not @scores[type][login].nil?
        puts "WARNING : '#{type}', '#{login}' has many entries in 'scores'"
        $stdout.flush
      end
      @scores[type][login] = {:old => score, :day => 0}
      
      @max_scores[type] = {:login => "nameless tee", :score => 0}  if @max_scores[type].nil?
      @max_scores[type] = {:login => login, :score => score}  if score > @max_scores[type][:score]
    end
    res.clear
  end

  def close
    @conn.close
  end
  
  def get_day_score(type, login)
    @scores[type] = {}  if @scores[type].nil?
    if @scores[type][login].nil?
      return 0
    else
      return @scores[type][login][:day]
    end
  end
  
  def get_score(type, login)
    @scores[type] = {}  if @scores[type].nil?
    if @scores[type][login].nil?
      return 0
    else
      return @scores[type][login][:old] + @scores[type][login][:day]
    end
  end
  
  def incr_score(type, login, to_add)
    @scores[type] = {}  if @scores[type].nil?
    if @scores[type][login].nil?
      @scores[type][login] = {:old => 0, :day => to_add}
      @conn.exec("INSERT INTO scores VALUES ($1, $2, $3)", [login, to_add.to_s, type])
    else
      @scores[type][login][:day] += to_add
      @conn.exec("UPDATE scores SET score = score + $2 WHERE login = $1 AND type = $3", [login, to_add.to_s, type])
    end
    
    @max_day_scores[type] = {:login => "nameless tee", :score => 0}  if @max_day_scores[type].nil?
    if get_day_score(type, login) > @max_day_scores[type][:score]
      @max_day_scores[type] = {:login => login, :score => get_day_score(type, login)}
    end
    
    @max_scores[type] = {:login => "nameless tee", :score => 0}  if @max_scores[type].nil?
    if get_score(type, login) > @max_scores[type][:score]
      @max_scores[type] = {:login => login, :score => get_score(type, login)}
    end
  end
  
  def get_max_day_score(type)
    @max_day_scores[type] = {:login => "nameless tee", :score => 0}  if @max_day_scores[type].nil?
    @max_day_scores[type]
  end
  
  def get_max_score(type)
    @max_scores[type] = {:login => "nameless tee", :score => 0}  if @max_scores[type].nil?
    @max_scores[type]
  end
end


Thread.abort_on_exception = true
mutex = Mutex.new

scores = nil
Thread.new do
  mutex.synchronize {scores = Scores.new("claret")}
end

server_socket = TCPServer.new("localhost", 8400)
at_exit {server_socket.close}
puts "Scores server started"
$stdout.flush

loop do
  Thread.start(server_socket.accept) do |socket|
    begin
      puts "#{socket} connection opened"
      $stdout.flush
      cmd_list = {
        "get_day_score" => Proc.new {|type, login| scores.get_day_score(type, login).to_s + "\n"},
        "get_score" => Proc.new {|type, login| scores.get_score(type, login).to_s + "\n"},
        "incr_score" => Proc.new do |type, login, to_add|
          scores.incr_score(type, login, to_add.to_i)
          next ""
        end,
        "get_max_day_score" => Proc.new do |type|
          max = scores.get_max_day_score(type)
          "#{max[:login]}\n#{max[:score].to_s}\n"
        end,
        "get_max_score" => Proc.new do |type|
          max = scores.get_max_score(type)
          "#{max[:login]}\n#{max[:score].to_s}\n"
        end
      }
      loop do
        cmd = socket.gets
        cmd.chomp!
        if cmd_list[cmd].nil?
          puts "#{cmd} : invalid command"
          $stdout.flush
        else
          args = []
          cmd_list[cmd].arity.times do
            arg = socket.gets
            args << arg.chomp
          end
          result = ""
          mutex.synchronize {result = cmd_list[cmd].call(*args)}
          socket.print result
        end
      end
    rescue Exception => e
      puts "#{e} exception ; #{socket} closed"
      $stdout.flush
      socket.close
    end
  end
end
